# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
import yaml
import sys
import traceback
import array
import cairo
import ftplib
import random
import string
# import math
import datetime
try:
    from gi import require_version
except ImportError:
    print("ERROR: pygobject is not installed.")
    raise
try:
    require_version("Gtk", "3.0")
except ValueError:
    print("ERROR: gtk3 development files are not installed.")
    raise
except AttributeError:
    print("pygobject version too old.")
    raise
from gi.repository import Gtk, Gdk
from gi.repository import GObject
from os.path import abspath
from os.path import dirname
from os.path import isfile
from os.path import join
from pkg_resources import resource_exists
from pkg_resources import resource_filename
from logging import getLogger
from logging import basicConfig
from logging import INFO
from six import iteritems
from six import StringIO
import qrcode
try:
    from synapselabels import __version__
    assert __version__
except ImportError:
    __version__ = "UNKNOWN"


log = getLogger('synapselabels')
dnd_targets = [('text/plain', Gtk.TargetFlags.OTHER_WIDGET, 0), ]


def get_path(name):
    """
    Returns the absolute path name of a resource file
    File in the current directory overrides file in package
    """
    path = None
    if isfile(name):
        path = abspath(name)
    elif resource_exists(__name__, name):
        res_filename = resource_filename(__name__, name)
        res_path = abspath(join(dirname(__file__), res_filename))
        if isfile(res_path):
            path = res_path
    if path:
        log.info("located path: %s", path)
    else:
        log.error("could not locate path for: %s", name)
    return path


label_config_file = get_path('labelconfig.yaml')
glade_file = get_path('labelsgui.glade')


def pil2cairo(im):
    """Transform a PIL Image into a Cairo ImageSurface."""

    assert sys.byteorder == 'little', 'We don\'t support big endian'
    if im.mode != 'RGBA':
        im = im.convert('RGBA')

    s = im.tobytes('raw', 'BGRA')
    a = array.array('B', s)
    dest = cairo.ImageSurface(cairo.FORMAT_ARGB32, im.size[0], im.size[1])
    ctx = cairo.Context(dest)
    non_premult_src_wo_alpha = cairo.ImageSurface.create_for_data(
        a, cairo.FORMAT_RGB24, im.size[0], im.size[1])
    non_premult_src_alpha = cairo.ImageSurface.create_for_data(
        a, cairo.FORMAT_ARGB32, im.size[0], im.size[1])
    ctx.set_source_surface(non_premult_src_wo_alpha)
    ctx.mask_surface(non_premult_src_alpha)
    return dest


def get_selected_data(treeview):
        # print("print_selected_label")
        (model, pathlist) = treeview.get_selection().get_selected_rows()
        for path in pathlist:
            tree_iter = model.get_iter(path)
            for index, column in enumerate(
                    treeview.get_columns()):
                return model.get_value(tree_iter, index)


def on_drag_data_get(treeview, context, selection, info, timestamp):
    """ Extract data from the source of the DnD operation.

    Serialize iterators of selected tasks in format
    <iter>,<iter>,...,<iter> and set it as parameter of DND """

    selected_data = get_selected_data(treeview)
    selection.set(selection.get_target(), 0, selected_data)


def set_treeview_columns(treeview, columns):
    for i in range(len(columns)):
        # cellrenderer to render the text
        cell = Gtk.CellRendererText(editable=False)
        # cell.set_property("editable", True)
        # the text in the first column should be in boldface
        # if i == 0:
        #    cell.props.weight_set = True
        #    cell.props.weight = Pango.Weight.BOLD
        # the column is created
        col = Gtk.TreeViewColumn(
            columns[i], cell, text=i)
        # and it is appended to the treeview
        treeview.append_column(col)


def set_treeview_from_clipboard(treeview, with_header=True):
    textData = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD).wait_for_text()
    listmodel = None
    if textData is not None:
        # print("clipboard text:", repr(textData))
        for line in textData.splitlines():
            # print("line:", repr(line))
            texts = line.replace(',', '\t').split('\t')
            if listmodel is None:
                for column in treeview.get_columns():
                    treeview.remove_column(column)
                args = [GObject.TYPE_STRING] * len(texts)
                # print(args)
                listmodel = Gtk.ListStore(*args)
                # listmodel = Gtk.ListStore(str, str, str)
                if with_header:
                    columns = texts
                else:
                    columns = [
                        "Column{}".format(index) for
                        index in
                        range(0, len(texts))
                        ]
                    # print("appending:", texts)
                    listmodel.append(texts)
            else:
                # print("appending:", texts)
                listmodel.append(texts)
        if listmodel is not None:
            treeview.set_model(listmodel)
            set_treeview_columns(treeview, columns)
    else:
        print("No text on the clipboard.")


def set_treeview_from_texts(treeview, texts):
    # treeview.set_headers_visible(False)
    listmodel = None
    if texts is not None:
        # print("texts:", repr(texts))
        for line in texts:
            # print("line:", repr(line))
            if listmodel is None:
                for column in treeview.get_columns():
                    treeview.remove_column(column)
                listmodel = Gtk.ListStore(GObject.TYPE_STRING)
                # print("appending:", line)
                listmodel.append([line])
            else:
                # print("appending:", line)
                listmodel.append([line])
        if listmodel is not None:
            # print("setting model:", listmodel)
            treeview.set_model(listmodel)
            set_treeview_columns(treeview, ['Params'])


def fill_combo(combo, data):
    store = Gtk.ListStore(GObject.TYPE_STRING)
    for line in data:
        store.append([line])
    combo.set_model(store)
    combo.set_active(0)
    # And here's the new stuff:
    cell = Gtk.CellRendererText()
    combo.pack_start(cell, True)
    combo.add_attribute(cell, "text", 0)


def get_treeview_column(treeview, x_pos):
    for column in treeview.get_columns():
        if column.get_x_offset() <= x_pos:
            if x_pos <= (column.get_x_offset() + column.get_width()):
                return column


def drag_data_received(
        widget, context, x_pos, y_pos, selection, info, timestamp):
    data = selection.get_data()
    print('drag_data_received data:', data)
    if data:
        column = get_treeview_column(widget, x_pos)
        if column:
            print(
                "old column title:", repr(column.get_title()),
                "new column title:", repr(data)
                )
            column.set_title(data)


class LabelsGui(object):
    """
    This is the E20 GUI
    """
    def __init__(self):
        """
        Build GUI
        """
        self.model = None
        # Build GUI from Glade file
        self.builder = Gtk.Builder()
        self.builder.add_from_file(glade_file)
        # Get objects
        go = self.builder.get_object
        self.window = go('applicationwindow1')
        self.window.set_title('Synapse Labels - v{}'.format(__version__))
        self.drawingarea = go('drawingarea')
        self.drawingarea.connect("draw", self.on_draw)
        self.treeview_data_table = go('treeviewDataTable')
        self.treeview_data_table.get_selection().connect(
            "changed",
            self.data_tree_selection_changed,
            "dummy"
            )
        self.treeview_data_table.enable_model_drag_dest(
            dnd_targets,
            Gdk.DragAction.DEFAULT | Gdk.DragAction.MOVE
            )
        self.treeview_data_table.connect(
            'drag-data-received',
            drag_data_received,
            )
        # self.combobox = go('comboboxtext1')
        # self.mac_num = go('labelMacNum')
        self.entry_date_code = go('entryDateCode')
        self.button_today = go('buttonToday')
        self.button_today.connect(
            "clicked",
            self.button_today_clicked,
            "today button data"
            )
        self.buttonPaste = go('buttonPaste')
        self.buttonPaste.connect(
            "clicked",
            self.button_paste_clicked,
            "cool button 1"
            )
        self.buttonPasteWithHeader = go('buttonPasteWithHeader')
        self.buttonPasteWithHeader.connect(
            "clicked",
            self.button_paste_with_header_clicked,
            "cool button 2"
            )
        self.buttonPrintSelectedLabels = go('buttonPrintSelectedLabels')
        self.buttonPrintSelectedLabels.connect(
            "clicked",
            self.print_selected_labels,
            )
        self.buttonPrintEveryLabel = go('buttonPrintEveryLabel')
        self.buttonPrintEveryLabel.connect(
            "clicked",
            self.print_every_label,
            )
        self.cb_printers = go('comboboxPrinters')
        self.cb_models = go('comboboxModels')
        self.cb_models.connect(
            "changed",
            self.cb_models_selection_changed,
            "dummy"
            )

        self.treeview_params = go('treeviewParams')
        self.treeview_params.get_selection().connect(
            "changed",
            self.params_tree_selection_changed,
            "dummy"
            )
        self.treeview_params.enable_model_drag_source(
            Gdk.ModifierType.BUTTON1_MASK,
            dnd_targets,
            Gdk.DragAction.DEFAULT | Gdk.DragAction.MOVE
            )
        self.treeview_params.connect('drag_data_get', on_drag_data_get)

        self.process_config()
        # Connect signals
        self.builder.connect_signals(self)
        self.cb_models_selection_changed(self.cb_models)
        # Configure interface
        self.window.connect('delete-event', lambda x, y: Gtk.main_quit())
        # GObject.timeout_add(1000, self.update_mac_pool)
        # Everything is ready
        self.window.show()

    def print_selected_labels(self, widget, data=None):
        self.print_labels(self.get_selected_labels(widget, data))

    def get_selected_labels(self, widget, data=None):
        # print("print_selected_label")
        selection = self.treeview_data_table.get_selection()
        # print('selection:', selection)
        (model, pathlist) = selection.get_selected_rows()
        labels = []
        for path in pathlist:
            # print('path:', path)
            tree_iter = model.get_iter(path)
            # print('tree_iter:', tree_iter)
            label_data = {}
            for param in self.get_parameters():
                value = None
                for index, column in enumerate(
                        self.treeview_data_table.get_columns()):
                    if column.get_title() == param:
                        value = model.get_value(
                            tree_iter,
                            index
                            )
                        break
                # print('value:', value)
                if value is None:
                    if param == 'date code':
                        entry_date_code_text = self.entry_date_code.get_text()
                        if len(entry_date_code_text) == 4:
                            value = entry_date_code_text
                if value:
                    label_data[param] = value
                else:
                    print(repr(param), 'parameter is missing')
                    return
            labels.append(label_data)
        return labels

    def get_label_from_row(self, row):
        label_data = {}
        for param in self.get_parameters():
            value = None
            for index, column in enumerate(
                    self.treeview_data_table.get_columns()):
                if column.get_title() == param:
                    value = row[index]
                    break
            # print('value:', value)
            if value:
                label_data[param] = value
            else:
                print(repr(param), 'parameter is missing')
                return
        return label_data

    def print_every_label(self, widget, data=None):
        labels = []
        for row in self.treeview_data_table.get_model():
            label_data = self.get_label_from_row(row)
            if label_data:
                labels.append(label_data)
        self.print_labels(labels)

    def process_config(self):
        with open(label_config_file, 'r') as stream:
            yaml_data = yaml.load(stream)
        # print('yaml_data:', repr(yaml_data))
        if 'label types' in yaml_data:
            self.label_types = yaml_data['label types']
            # print('label types:', repr(self.label_types))
        else:
            raise ValueError('"label types" is not found in config file')
        if 'printers' in yaml_data:
            self.printers = yaml_data['printers']
            # print('printers:', repr(printers))
            fill_combo(self.cb_printers, self.printers.keys())
        else:
            raise ValueError('"printers" is not found in config file')
        if 'models' in yaml_data:
            self.models = yaml_data['models']
            # print('models:', repr(self.models))
            model_names = []
            for model in self.models:
                for model_name in model:
                    model_names.append(model_name)
            fill_combo(self.cb_models, model_names)
        else:
            raise ValueError('"models" is not found in config file')
        if 'labels' in yaml_data:
            self.labels = yaml_data['labels']
        else:
            raise ValueError('"labels" is not found in config file')

    def button_paste_clicked(self, widget, data=None):
        # print("button_paste_clicked - %s" % data)
        set_treeview_from_clipboard(self.treeview_data_table, with_header=False)

    def button_paste_with_header_clicked(self, widget, data=None):
        # print("button_paste_with_header_clicked - %s" % data)
        set_treeview_from_clipboard(self.treeview_data_table, with_header=True)

    def data_tree_selection_changed(self, widget, data=None):
        # print("selection_changed - %s" % data)
        csr = widget.count_selected_rows()
        # print("count_selected_rows:", csr)
        self.buttonPrintSelectedLabels.set_sensitive(csr > 0)
        selected_labels = self.get_selected_labels(widget, data)
        if len(selected_labels) == 1:
            self.size_x, self.size_y = self.get_label_size()
            self.size_x /= 25.4
            self.size_y /= 25.4
            self.selected_label = selected_labels[0]
            # print('preview_label:', self.selected_label)
            self.selected_label_file = self.get_label_file(self.selected_label)
            # self.preview_label_file(self.get_label_file(label))
            self.drawingarea.set_size_request(
                300.0 * self.size_x,
                300.0 * self.size_y
                )
            self.drawingarea.queue_draw()
            # self.window.reshow_with_initial_size()
        else:
            self.drawingarea.set_size_request(-1, -1)
            self.selected_label = None
            self.label_file = None
            # self.window.reshow_with_initial_size()

    # def preview_label_file(self, label_file):
    #     print('preview_label_file', label_file)
    #     for line in label_file.splitlines():
    #         if line.startswith('T'):
    #             print ("preview", line)
    #             control, text = line.split(';', 1)
    #             control = control[2:]
    #             pos_x, pos_y, rotate, font, size = control.split(',')
    #             print ("text", pos_x, pos_y, rotate, font, size, text)

    def on_draw(self, wid, cr):
        if not hasattr(self, 'size_x'):
            return
        if not hasattr(self, 'size_y'):
            return
        # print('wid:', cr, dir(wid))
        # print('cr:', cr, dir(cr))

        # cr.set_source_rgb(0.6, 0.6, 0.6)

        # cr.rectangle(20, 20, 120, 80)
        # cr.rectangle(180, 20, 80, 80)
        # cr.fill()

        # cr.arc(330, 60, 40, 0, 2 * math.pi)
        # cr.fill()

        # cr.arc(90, 160, 40, math.pi / 4, math.pi)
        # cr.fill()

        # cr.translate(220, 180)
        # cr.scale(1, 0.7)
        # cr.arc(0, 0, 50, 0, 2 * math.pi)
        # cr.fill()

        # cr.set_source_rgb(0, 0, 0)
        # cr.move_to(110, 110)
        # cr.show_text('test1')
        # cr.move_to(50, 50)
        # cr.show_text('test2')
        # TBD
        # cr.scale(1 / 300.0, 1 / 300.0)
        cr.set_font_size(300.0 * 6.0 / 72)
        # frame it!
        cr.set_line_width(1)
        cr.rectangle(
            0,
            0,
            300 * self.size_x,
            300 * self.size_y,
            )
        cr.set_source_rgb(255, 255, 255)
        cr.fill()
        # cr.set_source_rgb(0, 0, 0)
        # cr.stroke()

        for line in self.selected_label_file.splitlines():
            if line.startswith('T'):
                # print("preview", line)
                control, text = line.split(';', 1)
                control = control[2:]
                pos_x, pos_y, rotate, font, size = control.split(',')
                # print("text", pos_x, pos_y, rotate, font, size, text)
                cr.move_to(300 * float(pos_x), 300 * float(pos_y))
                if size.startswith('pt'):
                    size = 300.0 * float(size[2:]) / 72.0
                    cr.set_font_size(size)
                if font == '3':
                    cr.select_font_face(
                        "Swiss 721",
                        cairo.FONT_SLANT_NORMAL,
                        cairo.FONT_WEIGHT_NORMAL,
                        )
                elif font == '5':
                    cr.select_font_face(
                        "Swiss 721",
                        cairo.FONT_SLANT_NORMAL,
                        cairo.FONT_WEIGHT_BOLD,
                        )
                cr.set_source_rgb(0, 0, 0)
                cr.show_text(text)
            elif line.startswith('G'):
                # print("preview", line)
                # G 0.05,0.64,0;R:1.9,0.33,0.0066666
                control, text = line.split(';', 1)
                control = control[2:]
                pos_x, pos_y, rotate = control.split(',')
                function, function_params = text.split(':', 1)
                if function == 'R':
                    width, height, thickness = function_params.split(',', 2)
                    cr.set_line_width(300 * float(thickness))
                    cr.rectangle(
                        300 * float(pos_x),
                        300 * float(pos_y),
                        300 * float(width),
                        300 * float(height),
                        )
                    cr.set_source_rgb(0, 0, 0)
                    cr.stroke()
                else:
                    print('unsupported line:', line)
            elif line.startswith('B'):
                # print("preview", line)
                # B 1.39,0.08,0,QRCODE+EL{error level}+MODEL2,.0133333;{qrcode}
                control, text = line.split(';', 1)
                control = control[2:]
                pos_x, pos_y, rotate, barcode, size = control.split(',', 4)
                barcode_opts = barcode.split('+')
                if barcode_opts[0] == 'QRCODE':
                    version = 1
                    error_correction = qrcode.constants.ERROR_CORRECT_L
                    for opt in barcode_opts:
                        if opt == 'QRCODE':
                            pass
                        elif opt == 'MODEL1':
                            version = 1
                        elif opt == 'MODEL2':
                            version = 2
                        elif opt == 'ELL':
                            error_correction = qrcode.constants.ERROR_CORRECT_L
                        elif opt == 'ELM':
                            error_correction = qrcode.constants.ERROR_CORRECT_M
                        elif opt == 'ELQ':
                            error_correction = qrcode.constants.ERROR_CORRECT_Q
                        elif opt == 'ELH':
                            error_correction = qrcode.constants.ERROR_CORRECT_H
                        else:
                            print('unsupported qrcode option:', opt)
                    box_size = int(0.5 + (300.0 * float(size)))
                    # print('box_size:', box_size)
                    qr = qrcode.QRCode(
                        version=version,
                        error_correction=error_correction,
                        box_size=box_size,
                        border=0,
                        )
                    qr.add_data(text)
                    qr.make(fit=True)
                    img = qr.make_image()
                    # print('img:', img)
                    # image = cairo.ImageSurface.create_from_png('logo.png')
                    cr.set_source_surface(
                        pil2cairo(img),  # image,
                        300.0 * float(pos_x),
                        300.0 * float(pos_y)
                        )
                    cr.paint()
                    # print('cr', dir(cr))
                else:
                    print('unsupported line:', line)
            else:
                print('unsupported line:', line)

    def preview_label_file(self, label_file):
        print('preview_label_file', label_file)
        for line in label_file.splitlines():
            if line.startswith('T'):
                print("preview", line)
                control, text = line.split(';', 1)
                control = control[2:]
                pos_x, pos_y, rotate, font, size = control.split(',')
                print("text", pos_x, pos_y, rotate, font, size, text)

    def params_tree_selection_changed(self, widget, data=None):
        # print("selection_changed - %s" % data)
        # csr = widget.count_selected_rows()
        # print("params_tree_selection_changed count_selected_rows:", csr)
        pass

    def get_label_name(self):
        # print('get_label_name')
        for model in self.models:
            # print('model:', model)
            for model_name, label_name in iteritems(model):
                # print('model name:', model_name, 'label name:', label_name)
                if model_name == self.model:
                    # print('label_name:', label_name)
                    return label_name

    def get_label(self):
        # print('get_label')
        label = self.labels[self.get_label_name()]
        # print('label:', label)
        return label

    def get_parameters(self):
        # print('get_parameters')
        try:
            parameters = self.get_label()['parameters']
            # print('parameters:', parameters)
            return parameters
        except:
            # silent fail, because it is possible to have a label without param
            return []

    def get_label_type(self):
        return self.label_types[self.get_label()['label type']]

    def get_label_size(self):
        label_type = self.get_label_type()
        return label_type['width'], label_type['height']

    def cb_models_selection_changed(self, widget, data=None):
        # print("cb_models_selection_changed - %s" % data)
        tree_iter = widget.get_active_iter()
        if tree_iter is not None:
            self.model = widget.get_model()[tree_iter][0]
            # print("Selected model:", self.model)
            set_treeview_from_texts(
                self.treeview_params,
                self.get_parameters()
                )

    def button_today_clicked(self, widget, data=None):
        # print("button_today_clicked - %s" % data)
        today = datetime.date.today()
        todays_week = today.strftime("%W")
        todays_year = today.strftime("%Y")
        todays_date_code = str(todays_year[-2:]) + str(todays_week)
        self.entry_date_code.set_text(todays_date_code)

    def get_label_file(self, label_params):
        for param in self.get_parameters():
            if param not in label_params:
                raise ValueError('label parameter %s is missing' % repr(param))

        label_format_params = {'model': self.model}
        label_name = self.get_label_name()
        if label_name == "#02 with meid":
            label_format_params['qrcode'] = (
                "#02:" +
                self.model + ":" +
                label_params['date code'] + ":" +
                label_params['ethernet mac address'] + ":" +
                label_params['snap mac address'] + ":" +
                label_params['wifi mac address'] + ":" +
                label_params['modem meid']
                )
            label_format_params['error level'] = 'H'
        elif label_name == "#02 without meid":
            label_format_params['qrcode'] = (
                "#02:" +
                self.model + ":" +
                label_params['date code'] + ":" +
                label_params['ethernet mac address'] + ":" +
                label_params['snap mac address'] + ":" +
                label_params['wifi mac address']
                )
            label_format_params['error level'] = 'H'
        elif label_name == "#08":
            label_format_params['qrcode'] = (
                "#08:" +
                self.model + ":" +
                label_params['date code'] + ":" +
                label_params['ethernet mac address last 6'][-6:] + ":" +
                label_params['snap node address'][-6:] + ":" +
                label_params['wifi mac address'] + ":" +
                label_params['modem meid'] + ":" +
                label_params['sim iccid'] + ":"
                )
            label_format_params['error level'] = (
                'Q' if len(label_format_params['qrcode']) >= 80 else 'H'
                )
        label_format_params.update(label_params)
        # print('label_format_params:', label_format_params)
        retval = ''
        for item in self.get_label()['items']:
            retval += item.format(**label_format_params) + '\n'
        return retval

    def print_label_file_to(self, label_file, ip_address):
        # create random file name
        filename = ''.join(random.choice(string.digits) for _ in range(9))
        print("random file name:", filename)
        ftp = None
        try:
            ftp = ftplib.FTP(ip_address)
            ftp.login("root", "1234")
            ftp.cwd("execute")
            ftp.storlines('STOR ' + filename, StringIO(label_file))
        except:
            print("FTP Error")
            traceback.print_exc(file=sys.stdout)
        finally:
            if ftp:
                ftp.close()

    def print_label_file(self, label_file):
        print('print_label_file:"""\n%s"""' % label_file)

        tree_iter = self.cb_printers.get_active_iter()
        if tree_iter is not None:
            selected_printer = self.cb_printers.get_model()[tree_iter][0]
            print('selected_printer:', selected_printer)
            for printer_name, printer_ip in iteritems(self.printers):
                if printer_name == selected_printer:
                    print('printer_ip:', printer_ip)
                    self.print_label_file_to(label_file, printer_ip)

    def print_label(self, label):
        self.print_label_file(self.get_label_file(label))

    def print_labels(self, labels):
        print('printing model:', self.model, 'labels:', labels)
        if labels is not None:
            for label in labels:
                self.print_label(label)


def synapse_labels():
    basicConfig(level=INFO)
    gui = LabelsGui()
    assert gui
    Gtk.main()


if __name__ == "__main__":
    synapse_labels()
